#ifdef MDARRAYS_DEBUG
    #define APD_DEBUG_COMPILE
#endif
#include <apd/apd.h>
#include <misclib/misclib.h>
#include <dan_math/equal.h>

// Where to send error messages
#define MDARRAYS_ERROR_STREAM                stderr

// Defaults for print style
#define DEFAULT_PRINT_DELIMITER              ' '
#define DEFAULT_PRINT_LINE_DELIMITER         '\n'

#define DEFAULT_PRINT_CHAR_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_PRINT_CHAR_WIDTH             3
#define DEFAULT_PRINT_CHAR_PRECISION         1
#define DEFAULT_PRINT_UCHAR_TYPE             MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_PRINT_UCHAR_WIDTH            3
#define DEFAULT_PRINT_BOOL_PRECISION         1
#define DEFAULT_PRINT_BOOL_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_PRINT_BOOL_WIDTH             3
#define DEFAULT_PRINT_UCHAR_PRECISION        1
#define DEFAULT_PRINT_INT_TYPE               MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_PRINT_INT_WIDTH              5
#define DEFAULT_PRINT_INT_PRECISION          1 
#define DEFAULT_PRINT_UINT_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_PRINT_UINT_WIDTH             5
#define DEFAULT_PRINT_UINT_PRECISION         1 
#define DEFAULT_PRINT_LONG_INT_TYPE          MDARRAYS_PRINTF_TYPE_ld
#define DEFAULT_PRINT_LONG_INT_WIDTH         9
#define DEFAULT_PRINT_LONG_INT_PRECISION     1
#define DEFAULT_PRINT_FLOAT_TYPE             MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_PRINT_FLOAT_WIDTH            8
#define DEFAULT_PRINT_FLOAT_PRECISION        3
#define DEFAULT_PRINT_DOUBLE_TYPE            MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_PRINT_DOUBLE_WIDTH           10
#define DEFAULT_PRINT_DOUBLE_PRECISION       6
#define DEFAULT_PRINT_LONG_DOUBLE_TYPE       MDARRAYS_PRINTF_TYPE_Lg
#define DEFAULT_PRINT_LONG_DOUBLE_WIDTH      14
#define DEFAULT_PRINT_LONG_DOUBLE_PRECISION  10
#define DEFAULT_PRINT_REAL_TYPE              MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_PRINT_REAL_WIDTH             10
#define DEFAULT_PRINT_REAL_PRECISION         6
#define DEFAULT_PRINT_GPOINTER_TYPE          MDARRAYS_PRINTF_TYPE_lX
#define DEFAULT_PRINT_GPOINTER_WIDTH         1
#define DEFAULT_PRINT_GPOINTER_PRECISION     1 /* is this used? */

// Defaults for save style
#define DEFAULT_SAVE_DELIMITER              ' '
#define DEFAULT_SAVE_LINE_DELIMITER         '\n'

#define DEFAULT_SAVE_CHAR_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_SAVE_CHAR_WIDTH             3
#define DEFAULT_SAVE_CHAR_PRECISION         1
#define DEFAULT_SAVE_UCHAR_TYPE             MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_SAVE_UCHAR_WIDTH            3
#define DEFAULT_SAVE_UCHAR_PRECISION        1
#define DEFAULT_SAVE_BOOL_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_SAVE_BOOL_WIDTH             3
#define DEFAULT_SAVE_BOOL_PRECISION         1
#define DEFAULT_SAVE_INT_TYPE               MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_SAVE_INT_WIDTH              5
#define DEFAULT_SAVE_INT_PRECISION          1 
#define DEFAULT_SAVE_UINT_TYPE              MDARRAYS_PRINTF_TYPE_d
#define DEFAULT_SAVE_UINT_WIDTH             5
#define DEFAULT_SAVE_UINT_PRECISION         1 
#define DEFAULT_SAVE_LONG_INT_TYPE          MDARRAYS_PRINTF_TYPE_ld
#define DEFAULT_SAVE_LONG_INT_WIDTH         9
#define DEFAULT_SAVE_LONG_INT_PRECISION     1
#define DEFAULT_SAVE_FLOAT_TYPE             MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_SAVE_FLOAT_WIDTH            9
#define DEFAULT_SAVE_FLOAT_PRECISION        5
#define DEFAULT_SAVE_DOUBLE_TYPE            MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_SAVE_DOUBLE_WIDTH           21
#define DEFAULT_SAVE_DOUBLE_PRECISION       17
#define DEFAULT_SAVE_LONG_DOUBLE_TYPE       MDARRAYS_PRINTF_TYPE_Lg
#define DEFAULT_SAVE_LONG_DOUBLE_WIDTH      30
#define DEFAULT_SAVE_LONG_DOUBLE_PRECISION  26
#define DEFAULT_SAVE_REAL_TYPE              MDARRAYS_PRINTF_TYPE_g
#define DEFAULT_SAVE_REAL_WIDTH             21
#define DEFAULT_SAVE_REAL_PRECISION         17
#define DEFAULT_SAVE_GPOINTER_TYPE          MDARRAYS_PRINTF_TYPE_lX
#define DEFAULT_SAVE_GPOINTER_WIDTH         1
#define DEFAULT_SAVE_GPOINTER_PRECISION     1 /* is this used? */

#define MDARRAYS_ASSERT(assertion, ...) \
    APD_ASSERT_ALWAYS_LIBRARY(MDARRAYS_ERROR_STREAM, cfile, cline, assertion, __VA_ARGS__)

/***********************************/
//   Array testing routines
/***********************************/
void array_tests(void);
void test(void);

void _check_bounds_internal(char *cfile, uint cline, 
                            void *array, mdarray_info_struct *mdarray_info, 
                            const char *array_name, uint line_number, 
                            uint access_index[MDARRAYS_MAX_DIMENSIONS]);
void * _mdarrays_get_nonmdarray_access_pointer(char *cfile, uint cline,
                                               void *array, uint type, uint type_size, 
                                               uint dimensions, int *sizes, int *element);

void * _mdarrays_allocate(char *cfile, uint cline,
                          uint dimensions, uint type, uint type_size, 
                          int *sizes, char *name);
void _mdarrays_print_1D_element(char *cfile, uint cline, FILE *file, void *array,
                                uint type, uint type_size, uint element, 
                                print_style_struct *print_style, bool delimiter_on);

// =======================================
// Misc internal functions
// =======================================
void _mdarrays_get_type_from_string(char *cfile, uint cline, char *type_string, 
                                    uint *type);
void * _mdarrays_get_access_pointer(char *cfile, uint cline, void *array, 
                                    mdarray_info_struct *mdarray_info, 
                                    char *array_name, uint access_index[MDARRAYS_MAX_DIMENSIONS]);
bool _mdarrays_compare_values(char *cfile, uint cline, void *value1, void *value2, uint type);

