#include <mdarrays.h>
#include <internal.h>
#include <ctype.h>
#include <errno.h>
#define APD_ADJUST_REPORT_THRESHOLD 0

// Load files filled with mdarray_save and mdarray_save_nonmdarray
// back into an mdarray or nonmdarray.  Note that these are non
// mix and match; you can't save a nonmdarray, then load to an
// mdarray, it won't work except for the 1D case.

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_load_recurse(char *cfile, uint cline,
                            FILE *file, void * array, uint dimensions, uint d_counter, 
                            uint type, uint type_size, uint *sizes)
{
int index;
int current_dimension_size;
long double **double_pointer;
long double value;
bool success_flag, eof_flag;
size_t fread_result;
void *write_here;

MDARRAYS_ASSERT(file != NULL);
MDARRAYS_ASSERT(array != NULL);

APD_PUSH_FUNCTION(4, "loading array @ %lX (dim=%d)", (unsigned long int)(array), dimensions);
MDARRAYS_ASSERT(dimensions > 0);
MDARRAYS_ASSERT(d_counter <= dimensions);
MDARRAYS_ASSERT(type_size > 0);
MDARRAYS_ASSERT(sizes != NULL);
current_dimension_size = sizes[d_counter-1];

if (d_counter == dimensions)
  {
    for (index = 0; index < current_dimension_size; index++)
      {
        if (type == MDARRAYS_GPOINTER_TYPE)
          {
            write_here = (void *)(&(((char *)array)[index * type_size]));
            if (index == 0) { MDARRAYS_ASSERT(array == write_here); }
            fread_result = fread(write_here, 1, type_size, file);
            MDARRAYS_ASSERT(fread_result == type_size, "error reading gpointer array from file.");
            APD_REPORT(5, "Read %d bytes @%lX", type_size, (unsigned long int)write_here);
            APD_REPORT_ACTION(5, misc_print_memory(stdout, write_here, type_size); );
          }
        else
          {
            misc_read_long_double_from_file(file, &value, &success_flag, 
                                            &eof_flag, NULL/*eol_flag*/, 
                                            true /*advance_after_read*/);
            //MDARRAYS_ASSERT(eof_flag == false, "Unexpected end of file");
            MDARRAYS_ASSERT(success_flag, "error during misc_read_long_double_from_file");
            // Point write_here
            write_here = (void *)(&(((char *)array)[index * type_size]));
            switch(type)
              {
                case MDARRAYS_CHAR_TYPE: *((char *)write_here) = (char)value; break;
                case MDARRAYS_UCHAR_TYPE: *((unsigned char *)write_here) = (unsigned char)value; break;
                case MDARRAYS_BOOL_TYPE: *((bool *)write_here) = (bool)value; break;
                case MDARRAYS_INT_TYPE: *((int *)write_here) = (int)value; break;
                case MDARRAYS_UINT_TYPE: *((uint *)write_here) = (uint)value; break;
                case MDARRAYS_LONG_INT_TYPE: *((long int *)write_here) = (long int)value; break;
                case MDARRAYS_FLOAT_TYPE: *((float *)write_here) = (float)value; break;
                case MDARRAYS_DOUBLE_TYPE: *((double *)write_here) = (double)value; break;
                case MDARRAYS_LONG_DOUBLE_TYPE: *((long double *)write_here) = (long double)value; break;
                case MDARRAYS_REAL_TYPE: *((real *)write_here) = (real)value; break;
                default: APD_DIE("fatal error: unknown value of parameter: type %d", type);
              }
            APD_REPORT(5, "array[%d] == %Lg", index, value);
          }
      }
  }
else
  {
    double_pointer = (long double **)array;
    for (index = 0; index < current_dimension_size; index++)
      {
        _mdarrays_load_recurse(cfile, cline, file, double_pointer[index], dimensions, 
                               d_counter+1, type, type_size, sizes);
      }
  }

APD_POP_FUNCTION();
} // end _mdarrays_load_recurse()

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void
_mdarrays_get_description(FILE *file, char *filename, char **description)
//<0> _mdarrays_get_description(file, description)
//<0>
//<0> Read in a single line from file starting with '#'.
//<0> If description is non-NULL, return a newly-allocated string
//<0> with the contents of that line in *description.
{
APD_PUSH_FUNCTION(4);
#define temp_size 1000
static char *temp_guts[temp_size];
char *temp = (char *)temp_guts;
char *r = fgets(temp, temp_size, file);
bool read_description = false;
APD_ASSERT_ALWAYS(r, "error reading line from %s", filename);
uint len = strlen(temp);
APD_ASSERT_ALWAYS(len > 0, "empty description string from %s", filename);
if (description)
  {
    if (temp[0] != '#') 
      {
        *description = NULL;
      }
    else
      {
        uint dlen = len-1; // don't include '#' character
        *description = malloc(sizeof(char) * dlen);
        APD_ASSERT_ALWAYS(*description, "error allocating description (out of memory?)");
        char *start = &(temp[1]);
        strncpy(*description, start, dlen-1);
        (*description)[dlen-1] = '\0';
        read_description = true;
      }
  }
if (!read_description) rewind(file); // we just read data not description
APD_POP_FUNCTION();
}
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_load(char *cfile, uint cline, char *filename, char **description, void * array)
{
FILE *file;
mdarray_info_struct *mdarray_info;

MDARRAYS_ASSERT(filename != NULL);
MDARRAYS_ASSERT(array != NULL);
APD_PUSH_FUNCTION(4, "loading array @%lX from file %s", (unsigned long int)array, filename);

// Extra info from array header info
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");

// If it's a GPOINTER, you're not just pointing to a gpointer,
// you're pointing to a structure or something.
if (mdarray_info->type == MDARRAYS_GPOINTER_TYPE)
  {
    // direct memory to disk - open file as binary
    file = fopen(filename, "rb");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for reading (binary mode)", filename);
  }
else
  {
    file = fopen(filename, "r");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for reading", filename);
  }

//<1> Call _mdarrays_get_description() to load description or skip it
_mdarrays_get_description(file, filename, description);

// Recurse 
_mdarrays_load_recurse(cfile, cline,
                       file, array, mdarray_info->dimensions, 1, mdarray_info->type,
                       mdarray_info->type_size, mdarray_info->sizes);

fclose(file);
APD_REPORT(4, "closed %s", filename);
APD_POP_FUNCTION();
} // end _mdarrays_load()
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void *
_mdarrays_load_2d_dynamic(char *cfile, uint cline, char *filename,
                          char *type_string, uint type_size, uint *rows, uint *columns,
                          char **description, uint *status)
//<0> _mdarrays_load_2d_dynamic(cfile, cline, filename, type_string,
//<0>                           type_size, *rows, *columns, **description, *status)
//<0>
//<0> Read in a 2d array of type type_string/type_size from filename into a
//<0> new array. Return the new array.
//<0> The number of rows and columns read will be returned in rows and columns.
//<0> If successful, set status 0, otherwise 1.
//<0> If description is non-NULL, allocate a string and read the first line of the
//<0> file into it.
{
APD_PUSH_FUNCTION(4, "Loading unknown size 2D array (type=%s, type_size=%d)",
	                 type_string, type_size);

FILE *file = fopen(filename, "r");
MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
APD_REPORT(4, "opened %s for reading", filename);

//<1> Call _mdarrays_get_description() to skip description line
_mdarrays_get_description(file, filename, NULL/*description*/);

*rows = 0;
*columns = 0;
uint row = 0;
uint column = 0;
bool eof_flag;
bool eol_flag;
bool success_flag;
long double value = 0.0L;

//<1> call misc_advance_to_next_number() to advance to start of first number
misc_advance_to_next_number(file, &eof_flag, &eol_flag);
if (eof_flag)
  {
    APD_REPORT_ALWAYS("Unexpected empty array");
    if (status != NULL) *status = 1;
    goto _mdarrays_load_2d_dynamic_END;
  }

//<1> while !eof_flag
while (! eof_flag)
  {
    //<2> call misc_read_long_double_from_file() to get this number
    misc_read_long_double_from_file(file, &value, &success_flag, &eof_flag, &eol_flag,
                                    false/*advance_after_read*/);
    column++;
    APD_REPORT(5, "Increment column to %d (eol=%d, eof=%d, value=%Lg)", 
               column, eol_flag, eof_flag, value);

    //<2> call misc_advance_to_next_number() to advance to start of next number
    misc_advance_to_next_number(file, &eof_flag, &eol_flag);
    APD_REPORT(5, "Advanced to next number (eol=%d, eof=%d)", eol_flag, eof_flag);

    //<2> Die if EOF before end of column
    if (eof_flag && *columns > 0 && column != *columns)
      {
        APD_REPORT_ALWAYS("EOF before end of column (row=%d, column=%d)", row, column);
        if (status != NULL) *status = 1;
        goto _mdarrays_load_2d_dynamic_END;
      }

    //<2> If this is the EOL on the first row, set the number of columns
    if (eol_flag && *columns == 0)
      {
        *columns = column;
        APD_REPORT(5, "Setting *columns to %d", *columns);
      }

    //<2> If column == *columns, this is the end of the row, start next
    //<2> row.
    if (column == *columns)
      {
        if (!eol_flag) 
          {
            APD_REPORT_ALWAYS("No EOL at expected location (row=%d, column=%d)", row, column);
            if (status != NULL) *status = 1;
            goto _mdarrays_load_2d_dynamic_END;
          }
        row++;
        APD_REPORT(5, "Incrementing row to %d", row);
        column = 0;
      }
  } // end while !eof_flag
*rows = row;

//<1> If EOL was never seen, this is a 1-row array
if (*rows == 0) 
  {
    *rows = 1;
    *columns = column;
  }

fclose(file);
APD_REPORT(5, "Total rows=%d, columns=%d", *rows, *columns);
  
//<1> call _mdarrays_get_type_from_string() to get type from type_string
uint type;
_mdarrays_get_type_from_string(cfile, cline, type_string, &type);
APD_REPORT(5, "type=%d", type);

//<1> call _mdarrays_allocate() to allocate a new array with the right
//<1> dimensions.
int sizes[2] = {*rows, *columns};
void *array 
  = _mdarrays_allocate(cfile, cline, 2/*dimensions*/, type, type_size, sizes, "loaded array");

//<1> call _mdarrays_load() to get dimensioned array from file
_mdarrays_load(cfile, cline, filename, description, array);

if (status != NULL) *status = 0;
_mdarrays_load_2d_dynamic_END:;
APD_POP_FUNCTION();
return array;
}

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_load_nonmdarray(char *cfile, uint cline, char *filename, void * array, uint type, 
                               uint type_size, uint dimensions, uint *sizes)
{
uint col;
uint elements;
void *write_here;
bool success_flag = false;
bool eof_flag = false;
long double value = 0.0L;
char *char_single_pointer;
uchar *uchar_single_pointer;
bool *bool_single_pointer;
int *int_single_pointer;
uint *uint_single_pointer;
long int *long_int_single_pointer;
float *float_single_pointer;
double *double_single_pointer;
long double *long_double_single_pointer;
real *real_single_pointer;
size_t fread_result;
FILE *file;

APD_PUSH_FUNCTION(4, "Loading nonmdarray @ %lX (dim=%d, type=%d, type_size=%d)",
	    (unsigned long int)(&array), dimensions, type, type_size);
MDARRAYS_ASSERT(dimensions > 0, "Fatal error: dimensions < 1");

if (type == MDARRAYS_GPOINTER_TYPE)
  {
    // direct memory to disk - open file as binary
    file = fopen(filename, "rb");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for reading (binary mode)", filename);
  }
else
  {
    file = fopen(filename, "r");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for reading", filename);
  }

// Get total number of elements in array
elements = 1;
for (col = 0; col < dimensions; col++)
  {
    MDARRAYS_ASSERT(sizes[col] > 0);
    elements = elements * sizes[col];
  }
APD_REPORT(5, "%d elements in array", elements);

// Read in all elements, in memory order
for (col = 0; col < elements; col++)
  {
    #define lra(vtype, vtype2) \
        vtype2 ## _single_pointer = (vtype *)array; \
        misc_read_long_double_from_file(file, &value, &success_flag, \
                                        &eof_flag, NULL/*eol_flag*/, \
                                        true /*advance_after_read*/); \
        /*MDARRAYS_ASSERT(eof_flag == false, "Unexpected end of file");*/ \
        MDARRAYS_ASSERT(success_flag, \
                          "error during misc_read_long_double_from_file: you don't have " \
                          "enough elements in the file you're reading from."); \
        vtype2 ## _single_pointer[col] = (vtype)value; \
        APD_REPORT(5, "array[%d]=%Lg", col, (long double)vtype2 ## _single_pointer[col]); \
        do {} while(0) /* swallow the semicolon */

    switch (type)
      {
        case MDARRAYS_CHAR_TYPE: lra(char, char); break;
        case MDARRAYS_UCHAR_TYPE: lra(uchar, uchar); break;
        case MDARRAYS_BOOL_TYPE: lra(bool, bool); break;
        case MDARRAYS_INT_TYPE: lra(int, int); break;
        case MDARRAYS_UINT_TYPE: lra(uint, uint); break;
        case MDARRAYS_LONG_INT_TYPE: lra(long int, long_int); break;
        case MDARRAYS_FLOAT_TYPE: lra(float, float); break;
        case MDARRAYS_DOUBLE_TYPE: lra(double, double); break;
        case MDARRAYS_LONG_DOUBLE_TYPE: lra(long double, long_double); break;
        case MDARRAYS_REAL_TYPE: lra(real, real); break;
        case MDARRAYS_GPOINTER_TYPE:
            write_here = (void *)(&(((char *)array)[col * type_size]));
            fread_result = fread(write_here, 1, type_size, file);
            MDARRAYS_ASSERT(fread_result == type_size,
                              "error reading gpointer array from file.");
            APD_REPORT(5, "Read %d bytes @%lX", type_size, (unsigned long int)write_here);
        break;
        default: APD_DIE("fatal error: unknown value of parameter: type %d", type);
      } // end switch (type)
  } // end for col
  
fclose(file);
APD_POP_FUNCTION();
} // end _mdarrays_load_nonmdarray()
