#include <mdarrays.h>
#include <internal.h>

#define PRINT_STYLE 0
#define SAVE_STYLE 1

// +++++++++++++++++++++++++++++++++++++++++++++++++
// Set the printing or saving parameters different than the defaults
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_set_style_guts(char *cfile, uint cline, 
                              uint mode, void *array, bool defaults, char delimiter,
                              char line_delimiter, uint type, uint width, uint precision)
{
mdarray_info_struct *mdarray_info;
print_style_struct *print_style = NULL;
APD_PUSH_FUNCTION(4, "defaults=%d, delimiter=%c, type=%d, width=%d, precision=%d",
                  defaults, delimiter, type, width, precision);
if (!defaults)
  {
    MDARRAYS_ASSERT(type >= MDARRAYS_PRINTF_TYPE_c && type <= MDARRAYS_PRINTF_TYPE_u);
    MDARRAYS_ASSERT(width < 100);
    MDARRAYS_ASSERT(precision < 100);
  }

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");
APD_REPORT(5, "examining array \"%s\"", mdarray_info->name);
switch (mode)
  {
    case MDARRAYS_PRINT_STYLE:
        print_style = &(mdarray_info->print_style);
        break;
    case MDARRAYS_SAVE_STYLE:
        print_style = &(mdarray_info->save_style);
        break;
    default:
        APD_DIE("unknown mode %d", mode);
  }

if (defaults)
  {
    switch (mode)
      {
        case MDARRAYS_PRINT_STYLE:
            print_style->delimiter           = DEFAULT_PRINT_DELIMITER;
            print_style->line_delimiter      = DEFAULT_PRINT_LINE_DELIMITER;

            switch (mdarray_info->type)
              {
                case MDARRAYS_CHAR_TYPE:
                    print_style->type        = DEFAULT_PRINT_CHAR_TYPE;
                    print_style->width       = DEFAULT_PRINT_CHAR_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_CHAR_PRECISION;
                    break;
                case MDARRAYS_UCHAR_TYPE:
                    print_style->type        = DEFAULT_PRINT_UCHAR_TYPE;
                    print_style->width       = DEFAULT_PRINT_UCHAR_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_UCHAR_PRECISION;
                    break;
                case MDARRAYS_BOOL_TYPE:
                    print_style->type        = DEFAULT_PRINT_BOOL_TYPE;
                    print_style->width       = DEFAULT_PRINT_BOOL_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_BOOL_PRECISION;
                    break;
                case MDARRAYS_INT_TYPE:
                    print_style->type        = DEFAULT_PRINT_INT_TYPE;
                    print_style->width       = DEFAULT_PRINT_INT_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_INT_PRECISION;
                    break;
                case MDARRAYS_UINT_TYPE:
                    print_style->type        = DEFAULT_PRINT_UINT_TYPE;
                    print_style->width       = DEFAULT_PRINT_UINT_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_UINT_PRECISION;
                    break;
                case MDARRAYS_LONG_INT_TYPE:
                    print_style->type        = DEFAULT_PRINT_LONG_INT_TYPE;
                    print_style->width       = DEFAULT_PRINT_LONG_INT_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_LONG_INT_PRECISION;
                    break;
                case MDARRAYS_FLOAT_TYPE:
                    print_style->type        = DEFAULT_PRINT_FLOAT_TYPE;
                    print_style->width       = DEFAULT_PRINT_FLOAT_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_FLOAT_PRECISION;
                    break;
                case MDARRAYS_DOUBLE_TYPE:
                    print_style->type        = DEFAULT_PRINT_DOUBLE_TYPE;
                    print_style->width       = DEFAULT_PRINT_DOUBLE_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_DOUBLE_PRECISION;
                    break;
                case MDARRAYS_LONG_DOUBLE_TYPE:
                    print_style->type        = DEFAULT_PRINT_LONG_DOUBLE_TYPE;
                    print_style->width       = DEFAULT_PRINT_LONG_DOUBLE_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_LONG_DOUBLE_PRECISION;
                    break;
                case MDARRAYS_REAL_TYPE:
                    print_style->type        = DEFAULT_PRINT_REAL_TYPE;
                    print_style->width       = DEFAULT_PRINT_REAL_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_REAL_PRECISION;
                    break;
                case MDARRAYS_GPOINTER_TYPE:
                    print_style->type        = DEFAULT_PRINT_GPOINTER_TYPE;
                    print_style->width       = DEFAULT_PRINT_GPOINTER_WIDTH;
                    print_style->precision   = DEFAULT_PRINT_GPOINTER_PRECISION;
                    break;

                default:
                    APD_DIE("unknown mdarray type: %d", mdarray_info->type);
              }
            break;
        case MDARRAYS_SAVE_STYLE:
            print_style->delimiter           = DEFAULT_SAVE_DELIMITER;
            print_style->line_delimiter      = DEFAULT_SAVE_LINE_DELIMITER;

            switch (mdarray_info->type)
              {
                case MDARRAYS_CHAR_TYPE:
                    print_style->type        = DEFAULT_SAVE_CHAR_TYPE;
                    print_style->width       = DEFAULT_SAVE_CHAR_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_CHAR_PRECISION;
                    break;
                case MDARRAYS_UCHAR_TYPE:
                    print_style->type        = DEFAULT_SAVE_UCHAR_TYPE;
                    print_style->width       = DEFAULT_SAVE_UCHAR_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_UCHAR_PRECISION;
                    break;
                case MDARRAYS_BOOL_TYPE:
                    print_style->type        = DEFAULT_SAVE_BOOL_TYPE;
                    print_style->width       = DEFAULT_SAVE_BOOL_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_BOOL_PRECISION;
                    break;
                case MDARRAYS_INT_TYPE:
                    print_style->type        = DEFAULT_SAVE_INT_TYPE;
                    print_style->width       = DEFAULT_SAVE_INT_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_INT_PRECISION;
                    break;
                case MDARRAYS_UINT_TYPE:
                    print_style->type        = DEFAULT_SAVE_UINT_TYPE;
                    print_style->width       = DEFAULT_SAVE_UINT_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_UINT_PRECISION;
                    break;
                case MDARRAYS_LONG_INT_TYPE:
                    print_style->type        = DEFAULT_SAVE_LONG_INT_TYPE;
                    print_style->width       = DEFAULT_SAVE_LONG_INT_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_LONG_INT_PRECISION;
                    break;
                case MDARRAYS_FLOAT_TYPE:
                    print_style->type        = DEFAULT_SAVE_FLOAT_TYPE;
                    print_style->width       = DEFAULT_SAVE_FLOAT_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_FLOAT_PRECISION;
                    break;
                case MDARRAYS_DOUBLE_TYPE:
                    print_style->type        = DEFAULT_SAVE_DOUBLE_TYPE;
                    print_style->width       = DEFAULT_SAVE_DOUBLE_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_DOUBLE_PRECISION;
                    break;
                case MDARRAYS_LONG_DOUBLE_TYPE:
                    print_style->type        = DEFAULT_SAVE_LONG_DOUBLE_TYPE;
                    print_style->width       = DEFAULT_SAVE_LONG_DOUBLE_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_LONG_DOUBLE_PRECISION;
                    break;
                case MDARRAYS_REAL_TYPE:
                    print_style->type        = DEFAULT_SAVE_REAL_TYPE;
                    print_style->width       = DEFAULT_SAVE_REAL_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_REAL_PRECISION;
                    break;
                case MDARRAYS_GPOINTER_TYPE:
                    print_style->type        = DEFAULT_SAVE_GPOINTER_TYPE;
                    print_style->width       = DEFAULT_SAVE_GPOINTER_WIDTH;
                    print_style->precision   = DEFAULT_SAVE_GPOINTER_PRECISION;
                    break;

                default:
                    APD_DIE("unknown mdarray type: %d", mdarray_info->type);
              }
          break;
      } // end switch mode
  }
else // if !defaults
  {
    print_style->delimiter = delimiter;
    print_style->line_delimiter = line_delimiter;
    print_style->type        = type;
    print_style->width       = width;
    print_style->precision   = precision;
  }

APD_POP_FUNCTION();
} // end _mdarrays_set_style_guts()

